require("utils.io.table")

ham = {}

-- indexed lookup
local edges = {}
local labels = {}
local nedges = {}
local visited = {}
-- sorted list
local vertices = {}

local function visit(n)
  visited[n] = visited[n] + 1
  for _n in pairs(edges[n]) do
    edges[_n][n] = edges[_n][n] - 1
    assert(edges[_n][n] >= 0)
  end
end

local function unvisit(n)
  visited[n] = visited[n] - 1
  assert(visited[n] >= 0)
  for _n in pairs(edges[n]) do
    edges[_n][n] = edges[_n][n] + 1
    assert(edges[_n][n] <= nedges[n])
  end
end

local function unvisit_all()
  for a, e in pairs(edges) do
    visited[a] = 0
    for b in pairs(e) do
      edges[a][b] = nedges[b]
    end
  end
end

local function sub_procedure_1(path)
  local n, uc
  local lv = path[#path]
  
  -- conditions for choosing the next vertex:
  -- 1. neighbour to the last vertex (lv) in the path
  for _n, _uc in pairs(edges[lv]) do
    -- 2. unvisited
    if visited[_n] == 0 then
      -- 3. has the fewest unvisited neighbours or
      -- 4. the smallest label
      if n == nil or uc > _uc or labels[n] > labels[_n] then
        n = _n
        uc = _uc
      end
    end
  end

  return n
end

local function procedure_1(path)
  -- next vertex
  local n = sub_procedure_1(path)
  if n == nil then
    return
  end
  table.insert(path, n)
  visit(n)
  -- repeat as long as we get an extended path
  procedure_1(path)
end

local function sub_procedure_2(path)
  local i, w, uc
  -- last vertex
  local lv = path[#path]
  -- v is a vertex in the path and is a neighbour of lv
  for _v in pairs(edges[lv]) do
    -- all neighbours of lv are in the path by construction
    local _i = table.find(path, _v)
    assert(_i)
    -- consider the successor vertex (i + 1) of v
    local _v1 = path[_i + 1]
    -- conditions for choosing w:
    -- 1. neighbour to v1
    for _w, _uc in pairs(edges[_v1]) do
      -- 2. unvisited
      if visited[_w] == 0 then
        -- 3. has the most unvisited neighbours
        -- 4. has the smallest label
        if w == nil or uc < _uc or labels[w] > labels[_w] then
          i = _i
          w = _w
          uc = _uc
        end
      end
    end
  end
  return w, i
end

local p2_ep = {}

local function procedure_2(path)
  local w, i = sub_procedure_2(path)
  if w == nil then
    return
  end
  -- reorder the path of visited vertices:
  -- p[1] -> p[i], p[#p] -> p[i + 1]
  -- p[1] -> p[i]
  for j = 1, i do
    table.insert(p2_ep, path[j])
  end
  -- p[#p] -> p[i + 1]
  for j = #path, i + 1, -1 do
    table.insert(p2_ep, path[j])
  end
  assert(table.dcount(p2_ep) == 0)
  
  table.clear(path)
  table.copy(p2_ep, path)
  table.clear(p2_ep)

  table.insert(path, w)
  visit(w)

  -- repeat this procedure as long as we obtain an extended path
  procedure_1(path)
  procedure_2(path)
end

local function sub_procedure_2b(path, sp)
  -- find v and w where:
  -- 1. v is in the sub-path path[1] -> path[#path - 2]
  -- 2. v has a neighbour w
  -- 3. w is unvisited (outside the path)
  
  -- iterate path
  for i = 1, #path do
    local v = path[i]
    -- iterate unvisited vertices (starting from the smallest label)
    for p = 1, #vertices do
      local w = vertices[p]
      if edges[v][w] and visited[w] == 0 then
        -- we found an unvisited neighbour w

        -- starting from w, look for a new path amongst the unvisited vertices
        table.insert(sp, w)
        visit(w)
        procedure_1(sp)

        -- trim the path from the right, if necessary, so that:
        -- 1. the last vertex (lv) has a neighbor on the sub-path path[1] => path[#path - 2]
        -- 2. vi+1 is a neighbor of vj+1
        while #sp > 0 do
          local lv = sp[#sp]
          for j = #path - 1, i + 1, -1 do
            local _vj = path[j]
            local vi1 = path[i + 1]
            local vj1 = path[j + 1]
            if edges[_vj][lv] and edges[vi1][vj1] then
              return i, j
            end
          end
          lv = table.remove(sp)
          unvisit(lv)
        end
      end
    end
  end
end

local p2b_sp = {}
local p2b_ep = {}

local function procedure_2b(path)
  local vi, vj = sub_procedure_2b(path, p2b_sp)
  if vj == nil then
    return
  end
  -- if successful, we obtain the extended path:
  -- path[1] -> path[i], sp[1] -> sp[#sp], path[j] -> path[i], path[j + 1] -> path[#path]
  -- path[1] -> path[i]
  for z = 1, vi do
    table.insert(p2b_ep, path[z])
  end
  -- sp[1] -> sp[#sp]
  for z = 1, #p2b_sp do
    table.insert(p2b_ep, p2b_sp[z])
  end
  -- path[j] -> path[i]
  for z = vj, vi + 1, -1 do
    table.insert(p2b_ep, path[z])
  end
  -- path[j + 1] -> path[#path]
  for z = vj + 1, #path do
    table.insert(p2b_ep, path[z])
  end
  table.clear(path)
  table.copy(p2b_ep, path)
  table.clear(p2b_sp)
  table.clear(p2b_ep)
  
  -- repeat this procedure as long as we obtain an extended path
  procedure_1(path)
  procedure_2(path)
  procedure_2b(path)
end

local function procedure_2c(path)
  table.reverse(path)
  procedure_2b(path)
end





local function index(graph)
  table.clear(edges)
  table.clear(vertices)
  table.clear(labels)
  table.clear(nedges)
  
  edges = table.copy(graph)
  -- edge count
  for a, e in pairs(edges) do
    nedges[a] = table.count(e)
  end
  unvisit_all()
  -- sorted vertices
  for n in pairs(edges) do
    table.insert(vertices, n)
  end
  -- descending order of degrees d(1) ≥ d(2) ≥  ... ≥ d(n)
  table.sort(vertices, function(y, z) return nedges[y] > nedges[z] end)
  -- label the vertices 1, 2, ..., n 
  for i, e in ipairs(vertices) do
    labels[e] = i
  end
end

local function find_paths(graph)
  index(graph)
  -- loop through all vertices
  local paths = {}
  local p = {}
  for _, s in ipairs(vertices) do
    -- begin new path
    table.insert(p, s)
    visit(s)
    
    procedure_1(p)
    procedure_2(p)
    if #p < #vertices then
      procedure_2b(p)
    end
    if #p < #vertices then
      procedure_2c(p)
    end
    
    local path = {}
    table.copy(p, path)
    table.insert(paths, path)
    
    -- clear the path
    while #p > 0 do
      local l = table.remove(p)
      unvisit(l)
    end
  end
  return paths
end

local function check_exceptional()
  -- get a list of all degree 1 vertices
  local d1 = {}
  for a, n in pairs(nedges) do
    if n == 1 then
      table.insert(d1, a)
    end
  end
  local h = {}
  if #d1 == 1 or #d1 == 2 then
    local p = d1[1]
    local h1 = {}
    local visited = {}
    local quit = false
    while not quit do
      quit = true
      table.insert(h1, p)
      visited[p] = true
      for _, v in ipairs(vertices) do
        if edges[p][v] and nedges[v] == 2 and not visited[v] and quit then
          p = v
          quit = false
        end
      end
    end
    table.insert(h, h1)
    if #d1 == 2 then
      local p = d1[2]
      local h2 = {}
      local quit = false
      while not quit do
        quit = true
        table.insert(h2, p)
        visited[p] = true
        for _, v in ipairs(vertices) do
          if edges[p][v] and nedges[v] == 2 and not visited[v] and quit then
            p = v
            quit = false
          end
        end
      end
      table.insert(h, h2)
    end
  end
  return h
end


local function concatenate(graph, path0, path1)
  local x, y
  local path = {}

  for i = #path1, 1, -1 do
    for j = #path0, 1, -1 do
      if graph[ path1[i] ][ path0[j] ] then
        local check = true
        for k = 1, i do
          for l = 1, j do
            if path1[k] == path0[l] then
              check = false
            end
          end
        end
        if check then
          x = j
          y = i
          break
        end
      end
    end
    if x and y then
      break
    end
  end
  if x and y then
    for i = 1, x, 1 do
      table.insert(path, path0[i])
    end
    for j = y, 1, -1 do
      table.insert(path, path1[j])
    end
  else
    table.copy(path0, path)
  end
  return path
end

--[[
 If the graph G has more than two vertices of degree 1, there can be no Hamiltonian tour.
 If the graph G has exactly two vertices of degree 1,
 then G cannot have a Hamiltonian circuit but could have a Hamiltonian tour.
 Suppose the graph G has exactly two vertices of degree 1, a and b.
 Find a path a = a1, a2, ..., ar using the procedure of Part I such that a2, ..., ar are vertices of degree 2.
 Similarly, find a path b = b1, b2, ..., bs such that b2, ..., bs are vertices of degree 2.
 Now let Ga be the graph obtained by deleting the vertices a1, a2, ..., ar-1, b1, b2, ..., bs from G
 and let Gb be the graph obtained by deleting the vertices a1, a2, ..., ar, b1, b2, ..., bs-1 from G.
 Use Parts I and II to find paths in the graphs Ga and Gb.
 If a path in Ga can be connected to a path in Gb to form a path in G,
 use Parts I and II to try and extend such a path to a Hamiltonian tour in G.
]]
local function find_paths_exceptional(graph, d1)
  local graph0 = {}
  table.copy(graph, graph0)
  for i, n in pairs(graph) do
    for j in pairs(n) do
      local check = false
      for k = 1, #d1[1] + 1 do
        if i == d1[1][k] then
          check = true
        end
      end
      for k = 1, #d1[2] do
        if i == d1[2][k] then
          check = true
        end
      end
      for k = 1, #d1[1] + 1 do
        if j == d1[1][k] then
          check = true
        end
      end
      for k = 1, #d1[2] do
        if j == d1[2][k] then
          check = true
        end
      end
      if check == true then
        graph0[i][j] = nil
        graph0[j][i] = nil
      end
    end
  end

  local temp_paths0 = find_paths(graph0)
  local paths0 = {}
  index(graph)
  for i, path0 in ipairs(temp_paths0) do
    for _, v in ipairs(path0) do
      visit(v)
    end
    procedure_1(path0)
    procedure_2(path0)
    procedure_2b(path0)
    procedure_2c(path0)
    table.insert(paths0, table.copy(path0))
    for _, v in ipairs(path0) do
      unvisit(v)
    end
  end

  local graph1 = {}
  table.copy(graph, graph1)
  for i, n in pairs(graph) do
    for j in pairs(n) do
      local check = false
      for k = 1, #d1[1] do
        if i == d1[1][k] then
          check = true
        end
      end
      for k = 1, #d1[2] + 1 do
        if i == d1[2][k] then
          check = true
        end
      end
      for k = 1, #d1[1] do
        if j == d1[1][k] then
          check = true
        end
      end
      for k = 1, #d1[2] + 1 do
        if j == d1[2][k] then
          check = true
        end
      end
      if check == true then
        graph1[i][j] = nil
        graph1[j][i] = nil
      end
    end
  end

  local temp_paths1 = find_paths(graph1)
  local paths1 = {}
  index(graph)
  for i, path1 in ipairs(temp_paths1) do
    for _, v in ipairs(path1) do
      visit(v)
    end
    procedure_1(path1)
    procedure_2(path1)
    procedure_2b(path1)
    procedure_2c(path1)
    table.insert(paths1, table.copy(path1))
    for _, v in ipairs(path1) do
      unvisit(v)
    end
  end


  local gs = table.count(graph)
  local paths = {}

  local found = false
  for i, p0 in ipairs(paths0) do
    for j, p1 in ipairs(paths1) do
      if found then
        break
      end
      local p0r = table.copy(p0)
      local p1r = table.copy(p1)
      table.reverse(p0r)
      table.reverse(p1r)
      
      local p = {}

      unvisit_all()
      p = concatenate(graph, p0, p1)
      for _, v in ipairs(p) do
        visit(v)
      end
      procedure_1(p)
      procedure_2(p)
      procedure_2b(p)
      procedure_2c(p)
      table.insert(paths, table.copy(p))
      if #p == gs then
        found = true
        break
      end

      unvisit_all()
      p = concatenate(graph, p0, p1r)
      for _, v in ipairs(p) do
        visit(v)
      end
      procedure_1(p)
      procedure_2(p)
      procedure_2b(p)
      procedure_2c(p)
      table.insert(paths, table.copy(p))
      if #p == gs then
        found = true
        break
      end

      unvisit_all()
      p = concatenate(graph, p0r, p1)
      for _, v in ipairs(p) do
        visit(v)
      end
      procedure_1(p)
      procedure_2(p)
      procedure_2b(p)
      procedure_2c(p)
      table.insert(paths, table.copy(p))
      if #p == gs then
        found = true
        break
      end

      unvisit_all()
      p = concatenate(graph, p0r, p1r)
      for _, v in ipairs(p) do
        visit(v)
      end
      procedure_1(p)
      procedure_2(p)
      procedure_2b(p)
      procedure_2c(p)
      table.insert(paths, table.copy(p))
      if #p == gs then
        found = true
        break
      end
    end
  end

  return paths
end


function ham.find_paths(graph)
  index(graph)
  local d1 = check_exceptional()
  local tour_ex = #d1 == 2
  if tour_ex then
    return find_paths_exceptional(graph, d1)
  else
    return find_paths(graph)
  end
end
